/*
    Copyright (C) 1996-2001  Ulric Eriksson <ulric@siag.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <locale.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "../config.h"
#include "cmalloc.h"
#include "common.h"

/*#define DEBUG
*/
#include "traceme.h"

/* ---
common.c

Device independent functions for all three programs.
--- */

/* ---
Figure out the names of directories
*/

char *libdir; /* where plugins are; default /usr/local/lib/siag */
char *datadir;	  /* where Scheme files are; default /usr/local/share/siag */
char *docdir;	  /* where online docs are; default /usr/local/doc/siag */
char *siag_basedir;	/* $HOME/.siag */
char *siag_tmpdir;	/* $HOME/.siag/tmp */
char *version;	  /* version string */

static struct {
	char *name;
	int value;
} cchar[] = {
	{"quot", '"'}, {"amp", '&'}, {"lt", '<'}, {"gt", '>'},
	{"nbsp", 160}, {"iexcl", 161}, {"cent", 162}, {"pound", 163},
	{"curren", 164}, {"yen", 165}, {"brvbar", 166}, {"sect", 167},
	{"uml", 168}, {"copy", 169}, {"ordf", 170}, {"laquo", 171},
	{"not", 172}, {"shy", 173}, {"reg", 174}, {"macr", 175},
	{"deg", 176}, {"plusmn", 177}, {"sup2", 178}, {"sup3", 179},
	{"acute", 180}, {"micro", 181}, {"para", 182}, {"middot", 183},
	{"cedil", 184}, {"sup1", 185}, {"ordm", 186}, {"raquo", 187},
	{"frac14", 188}, {"frac12", 189}, {"frac34", 190}, {"iquest", 191},
	{"Agrave", 192}, {"Aacute", 193}, {"Acirc", 194}, {"Atilde", 195},
	{"Auml", 196}, {"Aring", 197}, {"AElig", 198}, {"Ccedil", 199},
	{"Egrave", 200}, {"Eacute", 201}, {"Ecirc", 202}, {"Euml", 203},
	{"Igrave", 204}, {"Iacute", 205}, {"Icirc", 206}, {"Euml", 207},
	{"ETH", 208}, {"Ntilde", 209}, {"Ograve", 210}, {"Oacute", 211},
	{"Ocirc", 212}, {"Otilde", 213}, {"Ouml", 214}, {"times", 215},
	{"Oslash", 216}, {"Ugrave", 217}, {"Uacute", 218}, {"Ucirc", 219},
	{"Uuml", 220}, {"Yacute", 221}, {"THORN", 222}, {"szlig", 223},
	{"agrave", 224}, {"aacute", 225}, {"acirc", 226}, {"atilde", 227},
	{"auml", 228}, {"aring", 229}, {"aelig", 230}, {"ccedil", 231},
	{"egrave", 232}, {"eacute", 233}, {"ecirc", 234}, {"euml", 235},
	{"igrave", 236}, {"iacute", 237}, {"icirc", 238}, {"iuml", 239},
	{"eth", 240}, {"ntilde", 241}, {"ograve", 242}, {"oacute", 243},
	{"ocirc", 244}, {"otilde", 245}, {"ouml", 246}, {"divide", 247},
	{"slash", 248}, {"ugrave", 249}, {"uacute", 250}, {"ucirc", 251},
	{"uuml", 252}, {"yacute", 253}, {"thorn", 254}, {"yuml", 255},
	{NULL, 0}
};

/* convert Auml to  and #33 to ! */
int from_cchar(char *from)
{
	int i;

	if (from[0] == '#') {
		i = atoi(from+1);
		if (i >= ' ' && i <= 255) return i;
		return -1;
	}
	for (i = 0; cchar[i].name; i++) {
		if (!strcmp(cchar[i].name, from)) return cchar[i].value;
	}
	return -1;
}

/* convert  to &Auml; and so on */
void to_cchar(char *to, int from)
{
	int i;

	for (i = 0; cchar[i].value; i++) {
		if (from == cchar[i].value) {
			sprintf(to, "&%s;", cchar[i].name);
			return;
		}
	}
	to[0] = from;
	to[1] = '\0';
}

void common_init(char *fmt)
{
	char *home = getenv("SIAGHOME");
	char *docs = getenv("SIAGDOCS");
	char *lang;
	char p[1024];

	if (home) {
		libdir = datadir = cstrdup(home);
	} else {
		libdir = LIBDIR;
		datadir = DATADIR;
	}
	if (docs) {
		docdir = cstrdup(docs);
	} else {
		docdir = DOCDIR;
	}
	if (fmt) {
		sprintf(p, fmt, VERSION_NO);
	} else {
		sprintf(p, "Siag Office %s", VERSION_NO);
	}
	version = cstrdup(p);
	home = getenv("HOME");
	if (home == NULL) home = "";
	sprintf(p, "%s/.siag", home);
	siag_basedir = cstrdup(p);
	mkdir(siag_basedir, 0700);
	strcat(p, "/tmp");
	siag_tmpdir = cstrdup(p);
	mkdir(siag_tmpdir, 0700);

	/* i18n */
	lang = getenv("LC_ALL");
	if (lang == NULL) lang = getenv("LC_MESSAGES");
	if (lang == NULL) lang = getenv("LANG");
	if (lang) {
		sprintf(p, "%s/common/dictionary.%s", datadir, lang);
		load_dictionary(p);
	}
}

/* ---
Strip trailing newline, if there is one
*/

void chomp(char *p)
{
	if ((p = strchr(p, '\n'))) *p = '\0';
}

/* ---
Check if a string should be interpreted as a positive answer.
*/

int itsayes(char *p)
{
	return !cstrcasecmp(p, "yes")
		|| !cstrcasecmp(p, "1")
		|| !cstrcasecmp(p, "on")
		|| !cstrcasecmp(p, _("yes"))
		|| !cstrcasecmp(p, _("on"));
}

/* ---
strcmp which is immune to libc silliness
*/
int cstrcmp(const char *p, const char *q)
{
	int c;

	while (!(c = *p-*q) && *p) {
		p++;
		q++;
	}
	return c;
}

/* ---
Case insensitive compare
*/

int cstrcasecmp(const char *p, const char *q)
{
	int c;

	while (!(c = toupper(*p)-toupper(*q)) && *p) {
		p++;
		q++;
	}
	return c;
}

/* ---
Case insensitive compare, length limited
*/

int cstrncasecmp(const char *p, const char *q, size_t n)
{
	size_t i = 0;
	int c = 0;

	while ((i < n) && !(c = toupper(*p)-toupper(*q)) && *p) {
		p++;
		q++;
		i++;
	}
	return c;
}

int csnprintf(char *str, size_t n, const char *format, ...)
{
	int m;
	va_list ap;
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	m = vsnprintf(str, n, format, ap);
#else
	m = vsprintf(str, format, ap);
#endif
	va_end(ap);
	return m;
}


/* a little bookkeeping to keep track of temp files and make sure they
   are deleted when child processes terminate */

static int ndeletia;
static struct {
	long p;
	char *fn;
} *deletia;

/* ---
associate a temp file with a process
*/

void deletia_add(long p, char *fn)
{
	int i;

	for (i = 0; i < ndeletia; i++)
		if (deletia[i].fn == NULL) break;
	if (i == ndeletia) {
		ndeletia++;
		deletia = crealloc(deletia, ndeletia*sizeof *deletia);
	}
	deletia[i].p = p;
	deletia[i].fn = cstrdup(fn);
}

/* ---
mark all files related to a process to be deleted later
p == 0 marks all files
*/

void deletia_mark(long p)
{
	int i;

	for (i = 0; i < ndeletia; i++) {
		if (p == 0 || p == deletia[i].p)
			deletia[i].p = 0;
	}
}

/* ---
remove files that have previously been marked
*/

void deletia_reap(void)
{
	int i;

	for (i = 0; i < ndeletia; i++) {
		if (deletia[i].p == 0 && deletia[i].fn) {
			remove(deletia[i].fn);
			cfree(deletia[i].fn);
			deletia[i].fn = NULL;
		}
	}
}

/* ---
Signal handler for dead children.
*/

void waitforchild(int i)
{
	pid_t p;
	long lp;

	p = waitpid(-1, NULL, WNOHANG);
	lp = (long)p;
	if (p > 0) deletia_mark(lp);
	signal(SIGCHLD, waitforchild);
}

#define TAR_MAGIC "ustar"

/* ---
Check if a file is a tar archive and if an index file is present.
Returns 1 for success.
*/

int tryuntar(char *fn, char *index)
{
	FILE *fp;
	char b[1024];
	int result, n;
	struct stat statbuf;

	fp = fopen(fn, "r");
	if (!fp) {
		return 0;
	}
	n = fread(b, 1, 300, fp);
	if (n != 300) {
		fclose(fp);
		return 0;
	}
	if (strncmp(b+257, TAR_MAGIC, strlen(TAR_MAGIC))) {
		fclose(fp);
		return 0;
	}
	fclose(fp);
	sprintf(b,
		"mkdir -p %s/untar;"
		"cat %s 2> /dev/null |(cd %s/untar;tar xf - %s) 2> /dev/null",
		siag_basedir, fn, siag_basedir, index);
	system(b);
	sprintf(b, "%s/untar/%s", siag_basedir, index);
	result = !stat(b, &statbuf);
	sprintf(b, "rm -rf %s/untar", siag_basedir);
	system(b);
	return result;
}

/* ---
A function that doesn't do anything. Can be used to set break points
during debugging.
*/

void dummy_fun(void)
{
	;
}

static struct dictstruct {
	char *key, *xl;
} *dict = NULL;

static long nw = 0;

/* ---
Replace \n with real newlines and \t with real tabs.
Other escaped characters are replaced with themselves.
*/

static void unescape(char *p)
{
	char *q = p;
	int c, state = 0;
	while ((c = *p++)) {
		switch (state) {
		case 0:
			if (c == '\\') state = 1;
			else *q++ = c;
			break;
		case 1:
			if (c == 'n') *q++ = '\n';
			else if (c == 'r') *q++ = '\r';
			else if (c == 'b') *q++ = '\b';
			else if (c == 't') *q++ = '\t';
			else *q++ = c;
			state = 0;
			break;
		}
	}
	*q = '\0';
}

static int compar(const void *p, const void *q)
{
	struct dictstruct *p1 = (struct dictstruct *)p;
	struct dictstruct *q1 = (struct dictstruct *)q;
	return cstrcmp(p1->key, q1->key);
}

/* ---
Load the message string dictionary from $(SIAGHOME)/common/dictionary.$(LANG)
*/

void load_dictionary(char *fn)
{
        FILE *fp;
        char b[10000], *p, *q;

	/* convert everything after language code to upper case */
	p = strchr(fn, '_');
	if (p) {
		q = p;
		while (*q) {
			*q = toupper(*q);
			q++;
		}
	}

	if (p) {
		q = strchr(p, '.');
	} else {
		q = NULL;
	}

	fp = fopen(fn, "r");
	
	if (!fp && q) {
		/* try abridged file name */
		*q = '\0';
		fp = fopen(fn, "r");
	}
	if (!fp && p) {
		/* try abridged file name */
		*p = '\0';
		fp = fopen(fn, "r");
	}
        if (!fp) {
#ifdef DEBUG
                fprintf(stderr, "Can't open dictionary\n");
#endif
                return;
        }
        while (fgets(b, sizeof b, fp)) {
		if (b[0] == '#') continue;
                chomp(b);
                p = strchr(b, '\t');
                if (!p) continue;
                *p++ = 0;
                p += strspn(p, "\t");
                dict = crealloc(dict, (nw+1)*(sizeof dict[0]));
		unescape(b);
		unescape(p);
                dict[nw].key = cstrdup(b);
                dict[nw++].xl = cstrdup(p);
        }
	qsort(dict, nw, sizeof dict[0], compar);
}

/* ---
Translate a message string. If no match is found, the key is returned.
By using English keys, an application can keep working even when the
dictionary is incomplete.
990113: tree search through sorted list
*/

char *translate(char *key)
{
	long lower = 0, upper = nw-1;
        long i;
	int d;

	while (lower <= upper) {
		i = (lower+upper)/2;
		d = cstrcmp(key, dict[i].key);
		if (d == 0) return dict[i].xl;
		else if (d > 0) lower = i+1;
		else upper = i-1;
	}
        return key;
}

/* copy from q to p, quoting any character in b */
void quotecpy(char *p, char *q, char *b)
{
	int c;

	while ((c = *q++)) {
		if (strchr(b, c)) *p++ = '\\';
		*p++ = c;
	}
	*p = '\0';
}

pid_t spawn(const char *command)
{
	char *argv[20], *p, cmd[1024];
	int argc = 0;
	pid_t pid;

	strncpy(cmd, command, sizeof cmd);

	for (p = strtok(cmd, " \t\r\n");
	     p && argc < 20;
	     p = strtok(NULL, " \t\r\n")) {
		argv[argc++] = p;
	}
	argv[argc] = NULL;
	pid = fork();
	if (!pid) {
		/* this is the child */

		/*close(2);*/	/* don't display the child's messages */

		execvp(argv[0], argv);
		exit(0);
	}
	return pid;
}

int only_space(char *p)
{
	while (*p && isspace(*p)) p++;
	return (*p == '\0');
}

char *plugin_basedir(char *dir, char *bufname)
{
	int i;
	pid_t pid = getpid();

	sprintf(dir, "%s/%ld/%s", siag_basedir, (long)pid, bufname);
	for (i = strlen(dir)-strlen(bufname); dir[i]; i++)
		if (!isalnum(dir[i])) dir[i] = '_';
	return dir;
}


/* This translates ps fonts in .pfb format to ASCII ps files. */

/* 2000-10-26 Snarfed from Groff by Ulric */

/* Binary bytes per output line. */
#define BYTES_PER_LINE (64/2)
#define HEX_DIGITS "0123456789abcdef"

#define error(p) { fprintf(stderr, "%s\n", p); return 1; }

int pfb2ps(FILE *fpin, FILE *fpout)
{
	if (fpin == NULL || fpout == NULL) error("no file");
	for (;;) {
		int type, c, i;
		long n;

		c = getc(fpin);
		if (c != 0x80) {
			/* first byte of packet not 0x80, treat as ascii */
			do {
				putc(c, fpout);
				c = getc(fpin);
			} while (c != EOF);
			return 0;
		}
		type = getc(fpin);
		if (type == 3) break;
		if (type != 1 && type != 2) error("bad packet type");
		n = 0;
		for (i = 0; i < 4; i++) {
			c = getc(fpin);
			if (c == EOF) error("end of file in packet header");
			n |= (long) c << (i << 3);
		}
		if (n < 0) error("negative packet length");
		if (type == 1) {
			while (--n >= 0) {
				c = getc(fpin);
				if (c == EOF)
					error("end of file in text packet");
				if (c == '\r') c = '\n';
				putc(c, fpout);
			}
			if (c != '\n') putc('\n', fpout);
		} else {
			int count = 0;
			while (--n >= 0) {
				c = getc(fpin);
				if (c == EOF)
					error("end of file in binary packet");
				if (count >= BYTES_PER_LINE) {
					putc('\n', fpout);
					count = 0;
				}
				count++;
				putc(HEX_DIGITS[(c >> 4) & 0xf], fpout);
				putc(HEX_DIGITS[c & 0xf], fpout);
			}
			putc('\n', fpout);
		}
	}
	return 0;
}
